using System;
using System.Collections;

namespace LambdaCalculus
{
	/// <summary>
	/// Application is a class that represents a Lambda Calculus expression consisting of two other Lambda Calculus expressions.
	/// According to the language specification, it is:
	/// &lt;Application&gt; := &lt;Expression&gt; &lt;Expression&gt;
	/// </summary>
	internal class Application : Expression
	{
    /// <summary>
    /// the Application's left component expression
    /// </summary>
    private Expression oLeft;

    /// <summary>
    /// the Application's right component expression
    /// </summary>
    private Expression oRight;

    /// <summary>
    /// Construct an Application object from two Expressions.
    /// </summary>
    /// <param name="oLeft">the Application's left component expression</param>
    /// <param name="oRight">the Application's right component expression</param>
		internal Application( Expression oLeft, Expression oRight )
		{
			this.oLeft = oLeft;
      this.oRight = oRight;
		}

    internal override Expression Replace(Identifier oIdentifier, Expression oExpression)
    {
      return new Application( oLeft.Replace( oIdentifier, oExpression ), oRight.Replace( oIdentifier, oExpression ) );
    }

    public override Expression Evaluate()
    {
      
      if (this.oLeft.GetType() == typeof(Function))
        return ((Function) this.oLeft).Apply(this.oRight);
      else if (this.oLeft.GetType() == typeof(Application))
        return new Application( this.oLeft.Evaluate(), this.oRight );
      else
        return new Application( this.oLeft, this.oRight.Evaluate() );
    }

    internal override string ToString(IdentifierHash oIdentifiersInUse)
    {
      string sFormatString;
      if (oRight.GetType() == typeof(Application))
      {
        sFormatString = @"{0} ({1})";
      }
      else
      {
        sFormatString = @"{0} {1}";
      }
      return String.Format( sFormatString, this.oLeft.ToString(oIdentifiersInUse), this.oRight.ToString(oIdentifiersInUse) );
    }

    public override int IdentifierCount
    {
      get
      {
        return oLeft.IdentifierCount + oRight.IdentifierCount;
      }
    }


    internal override string Abbreviate(IdentifierHash oIdentifiersInUse)
    {
      string sFunctionName = oNamesAndExpressionText[this.ToString()];
      if (sFunctionName != null)
        return sFunctionName;
      string sFormatString;
      string sLeftAbbreviation = this.oLeft.Abbreviate(oIdentifiersInUse);
      string sRightAbbreviation = this.oRight.Abbreviate(oIdentifiersInUse);
      if (this.oRight.GetType() != typeof(Application) || sRightAbbreviation.IndexOf(' ') < 0)
      {
        sFormatString = @"{0} {1}";
      }
      else
      {
        sFormatString = @"{0} ({1})";
      }
      return String.Format( sFormatString, sLeftAbbreviation, sRightAbbreviation );
    }

    internal override string GetAbbreviatedLatex(IdentifierHash oIdentifiersInUse)
    {
      string sFunctionName = oNamesAndExpressionText[this.ToString()];
      if (sFunctionName != null)
        return sFunctionName;
      string sFormatString;
      string sLeftAbbreviation = this.oLeft.GetAbbreviatedLatex(oIdentifiersInUse);
      string sRightAbbreviation = this.oRight.GetAbbreviatedLatex(oIdentifiersInUse);
      if (this.oRight.GetType() != typeof(Application) || sRightAbbreviation.IndexOf(' ') < 0)
      {
        sFormatString = @"{0} {1}";
      }
      else
      {
        sFormatString = @"{0} ({1})";
      }
      return String.Format( sFormatString, sLeftAbbreviation, sRightAbbreviation );
    }

    public override Expression Copy()
    {
      return new Application( oLeft.Copy(), oRight.Copy() );
    }

    public override int FunctionDepth
    {
      get
      {
        int nLeftDepth = this.oLeft.FunctionDepth;
        int nRightDepth = this.oRight.FunctionDepth;
        return nLeftDepth > nRightDepth ? nLeftDepth : nRightDepth;
      }
    }

    internal override bool SameAs(Expression oExpression, ArrayList oThisIdentifiers, ArrayList oExpressionIdentifiers)
    {
      if (oExpression.GetType() != typeof(Application))
        return false;
      else
        return this.oLeft.SameAs(((Application)oExpression).oLeft, oThisIdentifiers, oExpressionIdentifiers ) &&
          this.oRight.SameAs(((Application)oExpression).oRight, oThisIdentifiers, oExpressionIdentifiers );
    }

    public override int LambdaCount
    {
      get
      {
        return this.oLeft.LambdaCount + this.oRight.LambdaCount;
      }
    }

    internal override string GetFullLatex(IdentifierHash oIdentifiersInUse)
    {
      string sFormatString;
      if (oRight.GetType() == typeof(Application))
      {
        sFormatString = @"{0} ({1})";
      }
      else
      {
        sFormatString = @"{0} {1}";
      }
      return String.Format( sFormatString,
                            this.oLeft.GetFullLatex(oIdentifiersInUse),
                            this.oRight.GetFullLatex(oIdentifiersInUse) );
    }


	}
}
