using System;
using System.Collections;
using System.Text.RegularExpressions;

namespace LambdaCalculus
{
  /// <summary>
  /// Expression is an abstract class that represents the common attributes of any Lambda Calculus expression.
  /// According to the language specification, it is:
  /// &lt;Expression&gt; := &lt;Identifier&gt;|&lt;Function&gt;|&lt;Application&gt;
  /// </summary>
  public abstract class Expression
  {

    #region Regular Expressions
    
    /// <summary>
    /// a Regular Expression used for locating whitespace within a Lambda Calculus Expression
    /// </summary>
    private static Regex rWhitespace = new Regex(@"\s+");
    /// <summary>
    /// a Regular Expression used for locating a free variable within a Lambda Calculus Expression
    /// </summary>
    private static Regex rFreeVariable = new Regex(@"^\s*[a-z]\s*$");
    /// <summary>
    /// a Regular Expression used for locating a Named Expression within a Lambda Calculus Expression
    /// </summary>
    private static Regex rNamedExpression = new Regex(@"^\s*[^a-z0-9()[\]\. \t\r\n][^()[\]\. \t\r\n]*\s*$");
    /// <summary>
    /// a Regular Expression used for locating a Natural Number within a Lambda Calculus Expression
    /// </summary>
    private static Regex rNaturalNumber = new Regex(@"^\s*[0-9]+\s*$");
    /// <summary>
    /// a Regular Expression used for locating the first characters of a Function definition within a Lambda Calculus Expression
    /// </summary>
    private static Regex rFunctionHeader = new Regex(@"^\s*\\\s*([a-z][a-z ]*)\s*\.(.*)\s*");

    #endregion

    #region Data Members

    /// <summary>
    /// a trie containing named expressions
    /// </summary>
    private static ExpressionTrie oExpressions = new ExpressionTrie();
    /// <summary>
    /// a trie containing a on-to-one mapping of Lambda Calculus Expressions and their names
    /// </summary>
    protected static StringTrie oNamesAndExpressionText = new StringTrie();
    /// <summary>
    /// storage for the Expression's string value, so it does not need to be contructed more than once
    /// </summary>
    private string sString;

    #endregion

    #region External Static Interface

    /// <summary>
    /// Create an Expression from a string, using any definitions in the dictionary.  This method will throw an exception if it
    /// cannot interpret the argued string.
    /// </summary>
    /// <param name="sExpression">a string to parse as a Lambda Calculus Expression</param>
    /// <returns>a new Expression object</returns>
    public static Expression Parse(string sExpression)
    {
      try
      {
        if (ValidParentheses(sExpression))
          return Parse(rWhitespace.Replace( sExpression, " " ), new IdentifierHash() );
        else      
          throw new ApplicationException(String.Format("Expression not evaluable; invalid parentheses: {0}", sExpression));
      }
      catch (ApplicationException oEx)
      {
        throw new ApplicationException( String.Format("Invalid expression '{0}':\n{1}", sExpression, oEx.Message ));
      }
    }

    /// <summary>
    /// Create a Church numeral for an unsigned integer (natural number).
    /// </summary>
    /// <param name="nNaturalNumber">a natural number</param>
    /// <returns>the Church numeral corresponding to the natural number</returns>
    public static Expression CreateChurchNumeral(uint nNaturalNumber)
    {
      Identifier oFirstArgument = new Identifier();
      Identifier oSecondArgument = new Identifier();
      Expression oCurrentComposition = oSecondArgument;
      for (int i = 1; i <= nNaturalNumber; i++)
      {
        oCurrentComposition = new Application( oFirstArgument, oCurrentComposition );
      }
      return new Function( oFirstArgument, new Function( oSecondArgument, oCurrentComposition ) );
    }

    /// <summary>
    /// Define a named expression using an expression statement of the form:
    /// <i>&lt;Expression Name&gt; &lt;:=|.=&gt; &lt;Expression&gt;</i>
    /// Where <i>Expression Name</i> is a string consisting of one or more characters with the following restrictions:
    /// <list type="number">
    /// <item><description>No parentheses, square brackets, periods, whitespace, or backslashes</description></item>
    /// <item><description>The first character may not be a lowercase letter or a decimal digit</description></item>
    /// </list>
    /// </summary>
    /// <param name="sDefinition">a definition statement</param>
    public static void Define(string sDefinition)
    {
      string[] oSplit = Regex.Split( sDefinition, @"\s*[:.]=\s*" );

      if (oSplit.Length != 2)
        throw new ApplicationException(String.Format( "Invalid dictionary definition: {0}", sDefinition ));
      if (!rNamedExpression.IsMatch(oSplit[0]))
        throw new ApplicationException(String.Format( "Invalid function name in dictionary definition: {0}", oSplit[0] ));

      Expression oDefinition = Expression.Parse(oSplit[1]);
      if (sDefinition.IndexOf(@":=") >= 0)
        oDefinition = oDefinition.Normalize();

      Define( oSplit[0], oDefinition );
    }

    /// <summary>
    /// Define a set of named expressions using an array of expression statements.
    /// </summary>
    /// <param name="oDefinitions">an array of expression statements</param>
    public static void Define(string[] oDefinitions)
    {
      foreach (string sDefinition in oDefinitions)
      {
        Define(sDefinition);
      }
    }

    /// <summary>
    /// Define a named expression.
    /// </summary>
    /// <param name="sName">the name of the expression</param>
    /// <param name="oExpression">the expression to be named</param>
    public static void Define( string sName, Expression oExpression )
    {
      // Place the Expression in the Expression dictionary under the function name.
      oExpressions[sName] = oExpression;
      // Place the Name in the Name dictionary under the Expression's string representation.
      oNamesAndExpressionText[oExpression.ToString()] = sName;
      // Place the Expression's string representation in the Name dictionary under the Expression's name.
      oNamesAndExpressionText[sName] = oExpression.ToString();
    }

    /// <summary>
    /// Undefine a named expression.
    /// </summary>
    /// <param name="sName">the name of a previously defined expression</param>
    public static void Undefine(string sName)
    {
      string sPreviousExpression = oNamesAndExpressionText[sName];
      if (sPreviousExpression == null)
        return;
      oNamesAndExpressionText[sPreviousExpression] = null;
      oNamesAndExpressionText[sName] = null;
      oExpressions[sName] = null;
    }
    /// <summary>
    /// Clear out all named expressions.
    /// </summary>
    public static void ClearDefinitions()
    {
      oExpressions = new ExpressionTrie();
      oNamesAndExpressionText = new StringTrie();
    }

    /// <summary>
    /// Get a named expression, indexed by its name.  Set a definition to null
    /// to remove it.
    /// </summary>
    public static Expression GetDefinition(string sExpressionName)
    {
      return oExpressions[sExpressionName];
    }


    #endregion

    #region External Instance Interface

    /// <summary>
    /// Evaluate the expression once and return the resulting expression.
    /// </summary>
    /// <returns>the expression, evaluated once</returns>
    public abstract Expression Evaluate();
    /// <summary>
    /// Evaluate this Expression until it will not reduce farther, then return the resulting Expression.
    /// </summary>
    /// <returns>the Expression, Normalized</returns>
    public Expression Normalize()
    {
      Expression oCurrentExpression = this.Copy();
      Expression oPreviousExpression;
      do
      {
        oPreviousExpression = oCurrentExpression;
        oCurrentExpression = oCurrentExpression.Evaluate();
      } while (!oCurrentExpression.SameAs(oPreviousExpression));

      return oPreviousExpression;
    }

    /// <summary>
    /// Evaluate this Expression until it will not reduce farther, then return the resulting Expression.
    /// </summary>
    /// <param name="oNormalization">this Expression, the normalized Expression, and each Expression in between</param>
    /// <returns>the Expression, Normalized</returns>
    public Expression Normalize(out Expression[] oNormalization)
    {
      ArrayList oExpressions = new ArrayList();
      Expression oCurrentExpression = this;
      do
      {
        oExpressions.Add(oCurrentExpression);
        oCurrentExpression = oCurrentExpression.Evaluate();
      } while (!oCurrentExpression.SameAs((Expression) oExpressions[oExpressions.Count - 1]));

      oNormalization = MakeExpressionArray(oExpressions);
      return (Expression) oExpressions[oExpressions.Count - 1];
    }

    /// <summary>
    /// Make a copy of the Expression, but leave any free variables intact.  This method is needed to avoid creating
    /// self-referencing Expression object, thus preventing infinite loops.
    /// </summary>
    /// <returns>a copy of the Expression</returns>
    public abstract Expression Copy();
    /// <summary>
    /// Convert the Expression object into unabbreviated text.
    /// </summary>
    /// <returns>an unabbreviated text representation of the Lambda Expression</returns>
    public override string ToString()
    {
      if (this.sString == null)
      {
        if (this.FunctionDepth > 26)
          this.sString = "[too complex]";
        else
          this.sString = this.ToString(new IdentifierHash());
      }
      return this.sString;
    }

    /// <summary>
    /// Get the number of Lambda operators within this expression, which is the number of functions within this expression,
    /// including the expression itself.
    /// </summary>
    public abstract int LambdaCount { get; }    

    /// <summary>
    /// Count the number of identifiers in function bodies in the expression.
    /// </summary>
    public abstract int IdentifierCount
    {
      get;
    }

    /// <summary>
    /// A text version of the Expression, with any Functions defined in the dictionary replaced with function names.
    /// </summary>
    public string Abbreviation
    {
      get
      {
        return this.Abbreviate(new IdentifierHash());
      }
    }

    /// <summary>
    /// A LaTeX version of the Expression, with any Functions defined in the dictionary replaced with function names.
    /// </summary>
    public string LatexAbbreviation
    {
      get
      {
        return this.GetAbbreviatedLatex(new IdentifierHash());
      }
    }

    /// <summary>
    /// A LaTeX version of the Expression.
    /// </summary>
    public string FullLatex
    {
      get
      {
        if (this.FunctionDepth > 26)
          return "[too complex]";

        string sExpression = this.GetFullLatex(new IdentifierHash());
        if (this.GetType() == typeof(Function))
          return sExpression.Substring( 1, sExpression.Length - 2 );
        else
          return sExpression;
      }
    }

    /// <summary>
    /// Convert the Expression to text, with any Functions defined in the dictionary replaced with function names.
    /// </summary>
    /// <param name="oIdentifiersInUse">any identifiers assigned letters by parent Expressions of this Expression</param>
    /// <returns>an abbreviated text version of the expression</returns>
    internal abstract string Abbreviate(IdentifierHash oIdentifiersInUse);

    /// <summary>
    /// Get the greatest number of open variables within any subexpression of this Expression.
    /// </summary>
    public abstract int FunctionDepth { get; }

    /// <summary>
    /// Determine whether this Expression is identical in structure to another Expression.
    /// </summary>
    /// <param name="oExpression">the Expression this Expression is to be compared to</param>
    /// <returns>true if the Expressions are identical structure, false otherwise</returns>
    public bool SameAs(Expression oExpression)
    {
      return SameAs( oExpression, new ArrayList(), new ArrayList() );
    }


    #endregion

    #region Internal Static Methods

    /// <summary>
    /// Create an Expression from a string, using any definitions in the dictionary and the names assigned to any free
    /// variables in the expression by a parent expression.  This method will throw an exception if it
    /// cannot interpret the argued string.
    /// </summary>
    /// <param name="sExpression">a string to parse as a Lambda Calculus Expression</param>
    /// <param name="oFreeVariables">characters assigned to any free variables in the expression</param>
    /// <returns>a new Expression object, with the free variables incorporated as needed</returns>
    private static Expression Parse(string sExpression, IdentifierHash oFreeVariables)
    {
      Expression oReturnExpression;

      //If the expression is a free variable, return the identifier object.
      if (rFreeVariable.IsMatch(sExpression))
      {
        char cFreeVariable = RemoveWhitespace(rFreeVariable.Match(sExpression).Value)[0];
        if (oFreeVariables[cFreeVariable] == null)
          throw new ApplicationException(String.Format( "Expression contains undeclared variable {0}", cFreeVariable ));
        return oFreeVariables[cFreeVariable];
      }

      //If the expression is a natural number, convert it to a Church numeral.
      if (rNaturalNumber.IsMatch(sExpression))
      {
        return CreateChurchNumeral(Convert.ToUInt32(sExpression));
      }

      //If the expression is a named expression, get its object.
      if (rNamedExpression.IsMatch(sExpression))
      {
        string sName = RemoveWhitespace(rNamedExpression.Match(sExpression).Value);
        Expression oNamedFunction = oExpressions[sName];
        if (oNamedFunction == null)
          throw new ApplicationException(String.Format( "Expression contains undefined named function {0}", sName ));
        return oNamedFunction;
      }
      
      //If the expression is a function, create a function objects for each argument.
      if (rFunctionHeader.IsMatch(sExpression))
      {
        string sArguments = RemoveWhitespace(rFunctionHeader.Match(sExpression).Groups[1].Value);
        string sBody = rFunctionHeader.Match(sExpression).Groups[2].Value;
        // Free the variables!
        for (int i = sArguments.Length - 1; i >= 0; i--)
        {
          if (oFreeVariables[sArguments[i]] != null)
            throw new ApplicationException(String.Format( "Function call has a duplicate argument identifier {0}", sExpression ));
          oFreeVariables[sArguments[i]] = new Identifier();
        }
        // Create an expression for the innermost function.
        oReturnExpression = new Function( oFreeVariables[sArguments[sArguments.Length - 1]],
                                          Parse( sBody, oFreeVariables ) );
        // Put inner functions inside outer functions.
        for (int i = sArguments.Length - 2; i >= 0; i--)
        {
          oReturnExpression = new Function( oFreeVariables[sArguments[i]], oReturnExpression );
        }
        // Release the free variables, now that they are bound.
        for (int i = sArguments.Length - 1; i >= 0; i--)
        {
          oFreeVariables.Close(sArguments[i]);
        }
        return oReturnExpression;
      }

      //If the expression is a word but not a legitimate variable or function name, throw an exception.
      if (!rWhitespace.IsMatch(sExpression) && sExpression.IndexOf('(') < 0)
      {
        throw new ApplicationException(String.Format( "Could not parse '{0}'.", sExpression ));
      }

      //If the expression is a function call or series of expressions, call them in order from left to right.
      ArrayList oSubExpressions = new ArrayList();
      int nCloser = -1;
      int nOpener = 0;
      while(nOpener < sExpression.Length)
      {
        nOpener = FindOpeningParenthesis( sExpression, nCloser + 1 );
        foreach (string sItem in Regex.Split(sExpression.Substring( nCloser + 1, nOpener - nCloser - 1 ), @"\s+"))
        {
          if (sItem.Length > 0)
            oSubExpressions.Add(Parse( sItem, oFreeVariables ));
        }
        if (nOpener < sExpression.Length)
        {
          if ((nCloser = FindMatchingParenthesis( sExpression, nOpener )) == sExpression.Length)
            throw new ApplicationException(String.Format( "Unmatched parenthesis found in expression: {0}", sExpression ));
          oSubExpressions.Add(Parse( sExpression.Substring( nOpener + 1, nCloser - nOpener - 1 ), oFreeVariables ));
        }
      }

      if (oSubExpressions.Count == 0)
        throw new ApplicationException("Empty expression encountered.");

      oReturnExpression = (Expression) oSubExpressions[0];
      for (int i = 1; i < oSubExpressions.Count; i++)
      {
        oReturnExpression = new Application( oReturnExpression, (Expression) oSubExpressions[i] );
      }
      return oReturnExpression;
      
    }

    /// <summary>
    /// Find the least complex Expression in an array of Expressions, at or after a specified index.
    /// </summary>
    /// <param name="oExpressions">an array of Expressions</param>
    /// <param name="nStartingIndex">the first index whose complexity is to be measured</param>
    /// <returns>the index of the least complex Expression at or after the starting index</returns>
    private static int LeastComplex( ArrayList oExpressions, int nStartingIndex )
    {
      int nLeastComplexity = ((Expression) oExpressions[nStartingIndex]).LambdaCount;
      int nLeastComplex = nStartingIndex;
      int nCurrentComplexity;
      for (int i = nStartingIndex; i < oExpressions.Count; i++)
      {
        nCurrentComplexity = ((Expression) oExpressions[i]).LambdaCount;
        if (nCurrentComplexity < nLeastComplexity)
        {
          nLeastComplexity = nCurrentComplexity;
          nLeastComplex = i;
        }
      }
      return nLeastComplex;
    }

    /// <summary>
    /// Check that a Lambda Calculus Expression contains matching parentheses
    /// </summary>
    /// <param name="sExpression">a Lambda Calculus Expression</param>
    /// <returns>true if each parenthesis in the expression has its match, false otherwise</returns>
    private static bool ValidParentheses(string sExpression)
    {
      int nOpens = 0;
      int nCloses = 0;
      for (int i = 0; i < sExpression.Length; i++)
      {
        if (sExpression[i] == ')') nCloses++;
        if (nCloses > nOpens) return false;
        if (sExpression[i] == '(') nOpens++;
      }
      return nOpens == nCloses;
    }

    /// <summary>
    /// Remove all whitespace from a string.
    /// </summary>
    /// <param name="sString">a string</param>
    /// <returns>the string, with any whitspace characters removed</returns>
    private static string RemoveWhitespace(string sString)
    {
      return rWhitespace.Replace( sString, "" );
    }

    /// <summary>
    /// Find the matching closing parenthesis for an opening parenthesis in a Lambda Calculus Expression.
    /// </summary>
    /// <param name="sString">a Lambda Calculus Expression</param>
    /// <param name="nOpeningParenthesisIndex">the index of the opening parenthesis whose closing parenthesis is to be found</param>
    /// <returns>the index of the matching closing parenthesis, or the string length if no matching closing parenthesis is found</returns>
    internal static int FindMatchingParenthesis( string sString, int nOpeningParenthesisIndex )
    {
      int nCounter = 0;
      for (int i = nOpeningParenthesisIndex; i < sString.Length; i++)
      {
        if (sString[i] == '(')
          nCounter++;
        else if (sString[i] == ')')
          nCounter--;

        if (nCounter == 0) return i;
      }
      return sString.Length;
    }

    /// <summary>
    /// Find the next opening parenthesis at or after a specified index in a Lambda Calculus Expression.
    /// </summary>
    /// <param name="sString">a Lambda Calculus Expression</param>
    /// <param name="nStartingIndex">the index to starting looking for opening parentheses at</param>
    /// <returns>the index of the first opening parenthesis at or after the index, or the string length if no opening parenthesis is found at or after the index</returns>
    internal static int FindOpeningParenthesis( string sString, int nStartingIndex )
    {
      for (int i = nStartingIndex; i < sString.Length; i++)
      {
        if (sString[i] == '(')
        {
          return i;
        }
      }
      return sString.Length;
    }

    /// <summary>
    /// Create an Expression array and fill it with the elements in an ArrayList
    /// </summary>
    /// <param name="oExpressions">an ArrayList containing only Expressions</param>
    /// <returns>an array of Expressions</returns>
    private static Expression[] MakeExpressionArray(ArrayList oExpressions)
    {
      Expression[] oArray = new Expression[oExpressions.Count];
      for (int i = 0; i < oExpressions.Count; i++)
      {
        oArray[i] = (Expression) oExpressions[i];
      }
      return oArray;
    }


    #endregion

    #region Internal Instance Methods

    /// <summary>
    /// Within the expression, replace all instances of an identifier with another expression.
    /// </summary>
    /// <param name="oIdentifier">the identifier to be replaced</param>
    /// <param name="oExpression">the expression to take place of the identifier</param>
    /// <returns>the Expression, with all instances of the identifier replaced with another expression</returns>
    internal abstract Expression Replace( Identifier oIdentifier, Expression oExpression );  
    /// <summary>
    /// Determine whether this Expression is identical in structure to another Expression.
    /// </summary>
    /// <param name="oExpression">the Expression this Expression is to be compared to</param>
    /// <param name="oThisIdentifiers">open variables within this expression</param>
    /// <param name="oExpressionIdentifiers">open variables within other expression</param>
    /// <returns></returns>
    internal abstract bool SameAs(Expression oExpression, ArrayList oThisIdentifiers, ArrayList oExpressionIdentifiers);
    /// <summary>
    /// Convert the Expression object into unabbreviated text.
    /// </summary>
    /// <param name="oIdentifiersInUse">any identifiers assigned letters by parent Expressions of this Expression</param>
    /// <returns>an unabbreviated text representation of the Lambda Expression</returns>
    internal abstract string ToString(IdentifierHash oIdentifiersInUse);
    /// <summary>
    /// Convert the Expression object into unabbreviated LaTeX.
    /// </summary>
    /// <param name="oIdentifiersInUse">any identifiers assigned letters by parent Expressions of this Expression</param>
    /// <returns>an unabbreviated LaTeX representation of the Lambda Expression</returns>
    internal abstract string GetFullLatex(IdentifierHash oIdentifiersInUse);
    /// <summary>
    /// Convert the Expression object into LaTeX, making use of defined function names.
    /// </summary>
    /// <param name="oIdentifiersInUse">any identifiers assigned letters by parent Expressions of this Expression</param>
    /// <returns>an abbreviated LaTeX representation of the Lambda Expression</returns>
    internal abstract string GetAbbreviatedLatex(IdentifierHash oIdentifiersInUse);

    #endregion
  }



 
}
