using System;
using System.Collections;
using System.Text.RegularExpressions;

namespace LambdaCalculus
{
	/// <summary>
  /// Function is a class that represents a Lambda Calculus expression consisting of an identifier and an expression that is
  /// evaluated by replacing any instances of the identifier with an expression applied to the Function.
	/// </summary>
	internal class Function : Expression
	{
    /// <summary>
    /// a Regular Expression used for locating a Church numerals greater than one in a Lambda Expression
    /// </summary>
    private static Regex rChurchNumeralGreaterThanOne = new Regex(@"\(\\a\.\(\\b\.(a \()+a b\)\)");

    /// <summary>
    /// the function's single argument
    /// </summary>
    private Identifier oArgument;
    /// <summary>
    /// the function's body
    /// </summary>
    private Expression oBody;

    /// <summary>
    /// Construct a Function from an Identifier and an Expression.
    /// </summary>
    /// <param name="oArgument">the function's single argument</param>
    /// <param name="oBody">the function's body</param>
		internal Function( Identifier oArgument, Expression oBody )
		{
			this.oArgument = oArgument;
      this.oBody = oBody;
		}

    internal override Expression Replace(Identifier oIdentifier, Expression oExpression)
    {
      Expression oNewBody = this.oBody.Replace( oIdentifier, oExpression );
      if (this.oArgument == oIdentifier)
        return oNewBody;
      else
        return new Function( this.oArgument, oNewBody );
    }

    /// <summary>
    /// Apply an Expression to the Function by replacing all instance's of the function's argument within the function's
    /// body with the Expression, then returning the function's body.
    /// </summary>
    /// <param name="oArgument"></param>
    /// <returns></returns>
    internal Expression Apply(Expression oArgument)
    {
      return this.oBody.Replace( this.oArgument, oArgument );
    }

    public override int IdentifierCount
    {
      get
      {
        return oBody.IdentifierCount;
      }
    }

    public override Expression Evaluate()
    {
      return new Function( this.oArgument, this.oBody.Evaluate() );
    }

    internal override string ToString(IdentifierHash oIdentifiersInUse)
    {
      string sFunction = String.Format( @"(\{0}.{1})", this.oArgument.ToString(oIdentifiersInUse), this.oBody.ToString(oIdentifiersInUse) );
      oIdentifiersInUse.Close(this.oArgument);
      return sFunction;
    }

    internal override string Abbreviate(IdentifierHash oIdentifiersInUse)
    {
      string sChurchNumeral = AbbreviateChurchNumeral(this.ToString());
      if (sChurchNumeral != null)
        return sChurchNumeral;
      string sFunctionName = oNamesAndExpressionText[this.ToString()];
      if (sFunctionName != null)
        return sFunctionName;
      string sFunction = String.Format( @"(\{0}.{1})",
        this.oArgument.Abbreviate(oIdentifiersInUse),
        this.oBody.Abbreviate(oIdentifiersInUse) );
      oIdentifiersInUse.Close(this.oArgument);
      return sFunction;
    }

    internal override string GetAbbreviatedLatex(IdentifierHash oIdentifiersInUse)
    {
      string sChurchNumeral = AbbreviateChurchNumeral(this.ToString());
      if (sChurchNumeral != null)
        return sChurchNumeral;
      string sFunctionName = oNamesAndExpressionText[this.ToString()];
      if (sFunctionName != null)
        return sFunctionName;
      string sArgument = this.oArgument.GetAbbreviatedLatex(oIdentifiersInUse);
      string sBody = this.oBody.GetAbbreviatedLatex(oIdentifiersInUse);
      oIdentifiersInUse.Close(this.oArgument);

      if (sBody.StartsWith(@"(\lambda "))
        return String.Format( @"(\lambda {0}{1}", sArgument, sBody.Substring(9) );
      else
        return String.Format( @"(\lambda {0}.{1})", sArgument, sBody );
    }

    private static string AbbreviateChurchNumeral(string sExpression)
    {
      if (sExpression == @"(\a.(\b.a b))")
        return "1";
      else if (rChurchNumeralGreaterThanOne.IsMatch(sExpression))
        return ((sExpression.Length - 9)/4).ToString();
      else
        return null;
    }

    public override Expression Copy()
    {
      Identifier oNewArgument = new Identifier();
      Expression oNewBody = this.oBody.Copy().Replace( this.oArgument, oNewArgument );
      return new Function( oNewArgument, oNewBody );
    }

    public override int FunctionDepth
    {
      get
      {
        return 1 + this.oBody.FunctionDepth;
      }
    }

    internal override bool SameAs(Expression oExpression, ArrayList oThisIdentifiers, ArrayList oExpressionIdentifiers)
    {
      if (oExpression.GetType() != typeof(Function))
      {
        return false;
      }
      else
      {
        oThisIdentifiers.Add(this.oArgument);
        oExpressionIdentifiers.Add(((Function) oExpression).oArgument);
        return this.oBody.SameAs(((Function) oExpression).oBody, oThisIdentifiers, oExpressionIdentifiers);
      }
    }

    public override int LambdaCount
    {
      get
      {
        return 1 + this.oBody.LambdaCount;
      }
    }

    internal override string GetFullLatex(IdentifierHash oIdentifiersInUse)
    {
      string sArgument = this.oArgument.GetFullLatex(oIdentifiersInUse);
      string sBody = this.oBody.GetFullLatex(oIdentifiersInUse);
      oIdentifiersInUse.Close(this.oArgument);

      if (sBody.StartsWith(@"(\lambda "))
        return String.Format( @"(\lambda {0}{1}", sArgument, sBody.Substring(9) );
      else
        return String.Format( @"(\lambda {0}.{1})", sArgument, sBody );
    }


	}
}
